from .errors import (
    RateLimited,
    Unauthorized,
    Forbidden,
    NotFound,
    BackblazeServerError,
    HTTPException
)
from .models import File, DeletedFile, DownloadedFile
from .file import LargeFile
from .bucket import Client

__all__ = (
    'RateLimited',
    'Unauthorized',
    'Forbidden',
    'NotFound',
    'BackblazeServerError',
    'HTTPException',
    'File',
    'DeletedFile',
    'DownloadedFile',
    'Client',
    'LargeFile'
)

__title__ = 'aiob2'
__author__ = 'Dan'
__license__ = 'MIT'
__version__ = '0.8.2'



from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Any

if TYPE_CHECKING:
    from aiohttp import ClientResponse

__all__ = ('HTTPException', 'RateLimited', 'Unauthorized', 'Forbidden', 'NotFound', 'BackblazeServerError')


class BackblazeException(Exception):
    """Base exception class for aiob2.

    Ideally speaking, this could be caught to handle any exceptions raised from this library.
    """

    pass


class HTTPException(BackblazeException):
    """Exception that's raised when a HTTP request operation fails.

    Attributes
    ------------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request. This is an
        instance of :class:`aiohttp.ClientResponse`.
    status: :class:`int`
        The code code of the HTTP request.
    text: :class:`str`
        The Spotify specific error code for the failure.
    """
    def __init__(self, response: ClientResponse, message: Dict[str, Any]):
        self.response: ClientResponse = response
        self.status: int = response.status
        self.code: str = message['code']
        self.text: str = message['message']

        fmt = '{0.status} (error code: {1}) {0.reason}: {2}'

        super().__init__(fmt.format(self.response, self.code, self.text))


class RateLimited(HTTPException):
    """Exception that's raised for when status code 429 occurs.

    Subclass of :exc:`HTTPException`
    """

    pass


class Unauthorized(HTTPException):
    """Exception that's raised for when status code 401 occurs.

    Subclass of :exc:`HTTPException`
    """

    pass


class Forbidden(HTTPException):
    """Exception that's raised for when status code 403 occurs.

    Subclass of :exc:`HTTPException`
    """

    pass


class NotFound(HTTPException):
    """Exception that's raised for when status code 404 occurs.

    Subclass of :exc:`HTTPException`
    """

    pass


class BackblazeServerError(Exception):
    """Exception that's raised for when a 500 range status code occurs.

    Subclass of :exc:`HTTPException`.
    """

    pass



import datetime
import logging
from typing import Optional, Literal, Dict, List, Any

import aiohttp

from .models import (
    File,
    DeletedFile,
    DownloadedFile
)
from .http import HTTPClient
from .file import LargeFile
from .utils import setup_logging, MISSING

__all__ = ('Client', )


class Client:
    """Represents an aiob2 Client that makes requests to Backblaze's B2 API.

    Parameters
    ----------
    application_key_id: :class:`str`
        The application key id to use for authentication.
    application_key: :class:`str`
        The application key to use for authentication.
    session: Optional[:class:`aiohttp.ClientSession`]
        An optional session to pass, otherwise one will be lazily created.
    logging_handler: Optional[:class:`logging.LogHandler`]
        The log handler to use for the library's logger. If this is ``None``
        then the library will not set up anything logging related. Logging
        will still work if ``None`` is passed, though it is your responsibility
        to set it up.

        The default log handler if not provided is :class:`logging.StreamHandler`.
    log_formatter: :class:`logging.Formatter`
        The formatter to use with the given log handler. If not provided then it
        defaults to a colour based logging formatter (if available).
    log_level: :class:`int`
        The default log level for the library's logger. This is only applied if the
        ``log_handler`` parameter is not ``None``. Defaults to ``logging.INFO``.
    root_logger: :class:`bool`
        Whether to set up the root logger rather than the library logger.
        By default, only the library logger (``'aiob2'``) is set up. If this
        is set to ``True`` then the root logger is set up as well.

        Defaults to ``False``.
    """
    def __init__(
        self,
        application_key_id: str,
        application_key: str,
        *,
        session: Optional[aiohttp.ClientSession] = None,
        log_handler: Optional[logging.Handler] = MISSING,
        log_formatter: logging.Formatter = MISSING,
        log_level: int = MISSING,
        root_logger: bool = False
    ):
        self._http = HTTPClient(application_key_id, application_key, session)
        if log_handler is not None:
            setup_logging(
                handler=log_handler,
                formatter=log_formatter,
                level=log_level,
                root=root_logger
            )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        await self.close()

    async def close(self):
        # This is really only possible if a Client is instantiated and no request is ever made
        if isinstance(self._http._session, aiohttp.ClientSession):  # type: ignore
            await self._http._session.close()  # type: ignore

    # TODO add all possible options; overload for customer encryption?
    async def upload_file(
        self,
        *,
        file_name: str,
        content_bytes: bytes,
        bucket_id: str,
        content_type: Optional[str] = None,
        content_disposition: Optional[str] = None,
        content_language: Optional[List[str]] = None,
        expires: Optional[datetime.datetime] = None,
        content_encoding: Optional[List[Literal['gzip', 'compress', 'deflate', 'identity']]] = None,
        comments: Optional[Dict[str, str]] = None,
        upload_timestamp: Optional[datetime.datetime] = None,
        server_side_encryption: Optional[Literal['AES256']] = None
    ) -> File:
        """Uploads a file to a bucket.

        Parameters
        -----------
        file_name: :class:`str`
            The name of the file.
        content_bytes: :class:`bytes`
            The raw bytes of the file to be uploaded.
        bucket_id: :class:`str`
            The ID of the bucket to upload to.
        content_type: Optional[:class:`str`]
            The content type of the content_bytes, e.g. video/mp4. This should be the original media/content,
            and not the result of the encodings applied. This is specified in the ``content_encoding``.

            B2's list of content types/extensions can be found [here](https://www.backblaze.com/b2/docs/content-types.html)

            If not provided, it will be automatically detected by Backblaze, and upon it not being discoverable, it'll
            default to ``application/octet-stream``.
        content_disposition: Optional[:class:`str`]
            Indicates whether the content is displayed inline in the browser, or as an attachment, which is
            locally downloaded.

            More info: [MDN Docs](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition)
        content_language: Optional[List[:class:`str`]]
            The intended language(s) for the audience. By default, when this is not specified, it indicates
            that the content is for all language audiences.

            For example, ``['en']`` (English) or ``['en', 'da']`` (English and Danish).
        expires: Optional[:class:`datetime.datetime`]
            Indicates the date/time after which the content is considered expired. This does NOT automatically delete the
            file.
        content_encoding: Optional[List[Literal[``gzip``, ``compress``, ``deflate``, ``identity``]]]
            Lists any encodings that have been applied to the content, and in what order.
        comments: Optional[Dict[:class:`str`, :class:`str`]]
            A key-value pair of strings denoting any extra information to store as metadata with the file.
            The key will be ``quote_plus`` encoded.
        upload_timestamp: Optional[datetime.datetime]
            The upload timestamp to use, instead of now.

            .. note ::
                Your account must be authorized to use this by Backblaze support.
        server_side_encryption: Optional[Literal[`AES256`]]
            Specifying this will encrypt the data before storing it using
            [Server-Side Encryption with Backblaze-Managed Keys](https://backblaze.com/b2/docs/server_side_encryption.html)
            with the specified algorithm, currently only ``AES256``.

        Returns
        ---------
        :class:`File`
            The uploaded file.
        """

        data = await self._http.upload_file(
            content_bytes=content_bytes,
            content_type=content_type or 'b2/x-auto',
            file_name=file_name,
            bucket_id=bucket_id,
            content_disposition=content_disposition,
            content_language=content_language,
            expires=expires,
            content_encoding=content_encoding,
            comments=comments,
            upload_timestamp=upload_timestamp,
            server_side_encryption=server_side_encryption
        )
        return File(data)

    async def upload_large_file(
        self,
        bucket_id: str,
        file_name: str,
        content_type: Optional[str] = None,
        upload_timestamp: Optional[datetime.datetime] = None,
        comments: Optional[Dict[Any, Any]] = None
    ) -> LargeFile:
        """Creates a large file to upload parts/chunks to incrementally.

        Parameters
        ----------
        bucket_id: str
            The ID of the bucket to upload to.
        file_name: str
            The name of the file.
        content_type: str
            The content type of the file once every part is combined together.

            If not provided, it will be automatically detected by Backblaze, and upon it not being discoverable, it'll
            default to ``application/octet-stream``.
        upload_timestamp: Optional[datetime.datetime]
            The upload timestamp to use, instead of now.

            .. note ::
                Your account must be authorized to use this by Backblaze support.
        comments: Optional[Dict[Any, Any]]
            Key-value pairs denoting any extra information to store as metadata with the file.

            Unlike ``upload_file``, multiple k-v pairs may be provided of any JSON-compatible data type.
        """

        data = await self._http.start_large_file(
            bucket_id=bucket_id,
            file_name=file_name,
            content_type=content_type or 'b2/x-auto',
            upload_timestamp=upload_timestamp,
            comments=comments
        )
        return LargeFile(data, self._http)

    async def delete_file(
        self,
        file_name: str,
        file_id: str
    ) -> DeletedFile:
        """Deletes a file from a bucket.

        Parameters
        -----------
        file_name: :class:`str`
            The name of the file to delete.
        file_id: :class:`str`
            The id of the file to delete.

        Returns
        ---------
        :class:`DeletedFile`
            The deleted file.
        """

        data = await self._http.delete_file(file_name=file_name, file_id=file_id)
        return DeletedFile(data)

    async def download_file_by_id(
        self,
        file_id: str,
        *,
        content_disposition: Optional[str] = None,
        content_language: Optional[str] = None,
        expires: Optional[str] = None,
        cache_control: Optional[str] = None,
        content_encoding: Optional[str] = None,
        content_type: Optional[str] = None,
        server_side_encryption: Optional[str] = None
    ) -> DownloadedFile:
        """Downloads a file.

        Parameters
        -----------
        file_id: :class:`str`
            The file id of the file to be downloaded.
        content_disposition: Optional[:class:`str`]
            Overrides the current 'b2-content-disposition' specified when the file was uploaded.
        content_language: Optional[:class:`str`]
            Overrides the current 'b2-content-language' specified when the file was uploaded.
        expires: Optional[:class:`str`]
            Overrides the current 'b2-expires' specified when the file was uploaded.
        cache_control: Optional[:class:`str`]
            Overrides the current 'b2-cache-control' specified when the file was uploaded.
        content_encoding: Optional[:class:`str`]
            Overrides the current 'b2-content-encoding' specified when the file was uploaded.
        content_type: Optional[:class:`str`]
            Overrides the current 'Content-Type' specified when the file was uploaded.
        server_side_encryption: Optional[:class:`str`]
            This is requires if the file was uploaded and stored using Server-Side Encryption with
            Customer-Managed Keyts (SSE-C)

        Returns
        ---------
        :class:`DownloadedFile`
            The file requested.
        """

        data = await self._http.download_file_by_id(
            file_id=file_id,
            content_disposition=content_disposition,
            content_language=content_language,
            expires=expires,
            cache_control=cache_control,
            content_encoding=content_encoding,
            content_type=content_type,
            server_side_encryption=server_side_encryption
        )
        return DownloadedFile(data[0], data[1])  # type: ignore

    async def download_file_by_name(
        self,
        file_name: str,
        bucket_name: str,
        *,
        content_disposition: Optional[str] = None,
        content_language: Optional[str] = None,
        expires: Optional[str] = None,
        cache_control: Optional[str] = None,
        content_encoding: Optional[str] = None,
        content_type: Optional[str] = None,
        server_side_encryption: Optional[str] = None
    ) -> DownloadedFile:
        """Downloads a file.

        Parameters
        -----------
        file_name: :class:`str`
            The file name of the file to be downloaded.
        bucket_name: :class:`str`
            The bucket name of the file to be downloaded. This should only be specified if you have specified
            file_name and not file_id.
        content_disposition: Optional[:class:`str`]
            Overrides the current 'b2-content-disposition' specified when the file was uploaded.
        content_language: Optional[:class:`str`]
            Overrides the current 'b2-content-language' specified when the file was uploaded.
        expires: Optional[:class:`str`]
            Overrides the current 'b2-expires' specified when the file was uploaded.
        cache_control: Optional[:class:`str`]
            Overrides the current 'b2-cache-control' specified when the file was uploaded.
        content_encoding: Optional[:class:`str`]
            Overrides the current 'b2-content-encoding' specified when the file was uploaded.
        content_type: Optional[:class:`str`]
            Overrides the current 'Content-Type' specified when the file was uploaded.
        server_side_encryption: Optional[:class:`str`]
            This is requires if the file was uploaded and stored using Server-Side Encryption with
            Customer-Managed Keyts (SSE-C)

        Returns
        ---------
        :class:`DownloadedFile`
            The file requested.
        """

        data = await self._http.download_file_by_name(
            file_name=file_name,
            bucket_name=bucket_name,
            content_disposition=content_disposition,
            content_language=content_language,
            expires=expires,
            cache_control=cache_control,
            content_encoding=content_encoding,
            content_type=content_type,
            server_side_encryption=server_side_encryption
        )
        return DownloadedFile(data[0], data[1])  # type: ignore



import datetime
import json
import logging
import sys
import os
from typing import Any

try:
    import orjson
except ModuleNotFoundError:
    HAS_ORJSON = False  # type: ignore
else:
    HAS_ORJSON = True

__all__ = (
    'format_timestamp',
    '_to_json',
    '_from_json',
    'MISSING',
    'setup_logging'
)


def format_timestamp(timestamp: float) -> datetime.datetime:
    """Formats a UTC float timestamp represented in milliseconds.

    Parameters
    ----------
    timestamp: :class:`float`
        The UTC timestamp in milliseconds.
    """
    # shifts the decimal point 3 digits to the left
    timestamp /= 1000.

    return datetime.datetime.utcfromtimestamp(timestamp)


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other: Any) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()

if HAS_ORJSON:

    def _to_json(obj: Any) -> str:
        return orjson.dumps(obj).decode('utf-8')

    _from_json = orjson.loads  # type: ignore

else:

    def _to_json(obj: Any) -> str:
        return json.dumps(obj, separators=(',', ':'), ensure_ascii=True)

    _from_json = json.loads


def is_docker() -> bool:
    path = '/proc/self/cgroup'
    return os.path.exists('/.dockerenv') or (os.path.isfile(path) and any('docker' in line for line in open(path)))


def stream_supports_colour(stream: Any) -> bool:
    # Pycharm and Vscode support colour in their inbuilt editors
    if 'PYCHARM_HOSTED' in os.environ or os.environ.get('TERM_PROGRAM') == 'vscode':
        return True

    is_a_tty = hasattr(stream, 'isatty') and stream.isatty()
    if sys.platform != 'win32':
        # Docker does not consistently have a tty attached to it
        return is_a_tty or is_docker()

    # ANSICON checks for things like ConEmu
    # WT_SESSION checks if this is Windows Terminal
    return is_a_tty and ('ANSICON' in os.environ or 'WT_SESSION' in os.environ)


class _ColourFormatter(logging.Formatter):

    # ANSI codes are a bit weird to decipher if you're unfamiliar with them, so here's a refresher
    # It starts off with a format like \x1b[XXXm where XXX is a semicolon separated list of commands
    # The important ones here relate to colour.
    # 30-37 are black, red, green, yellow, blue, magenta, cyan and white in that order
    # 40-47 are the same except for the background
    # 90-97 are the same but "bright" foreground
    # 100-107 are the same as the bright ones but for the background.
    # 1 means bold, 2 means dim, 0 means reset, and 4 means underline.

    LEVEL_COLOURS = [
        (logging.DEBUG, '\x1b[40;1m'),
        (logging.INFO, '\x1b[34;1m'),
        (logging.WARNING, '\x1b[33;1m'),
        (logging.ERROR, '\x1b[31m'),
        (logging.CRITICAL, '\x1b[41m'),
    ]

    FORMATS = {
        level: logging.Formatter(
            f'\x1b[30;1m%(asctime)s\x1b[0m {colour}%(levelname)-8s\x1b[0m \x1b[35m%(name)s\x1b[0m %(message)s',
            '%Y-%m-%d %H:%M:%S',
        )
        for level, colour in LEVEL_COLOURS
    }

    def format(self, record: Any):
        formatter = self.FORMATS.get(record.levelno)
        if formatter is None:
            formatter = self.FORMATS[logging.DEBUG]

        # Override the traceback to always print in red
        if record.exc_info:
            text = formatter.formatException(record.exc_info)
            record.exc_text = f'\x1b[31m{text}\x1b[0m'

        output = formatter.format(record)

        # Remove the cache layer
        record.exc_text = None
        return output


def setup_logging(
    *,
    handler: logging.Handler = MISSING,
    formatter: logging.Formatter = MISSING,
    level: int = MISSING,
    root: bool = True,
) -> None:
    """A helper function to setup logging.

    This is superficially similar to :func:`logging.basicConfig` but
    uses different defaults and a colour formatter if the stream can
    display colour.

    Parameters
    -----------
    handler: :class:`logging.Handler`
        The log handler to use for the library's logger.

        The default log handler if not provided is :class:`logging.StreamHandler`.
    formatter: :class:`logging.Formatter`
        The formatter to use with the given log handler. If not provided then it
        defaults to a colour based logging formatter (if available). If colour
        is not available then a simple logging formatter is provided.
    level: :class:`int`
        The default log level for the library's logger. Defaults to ``logging.INFO``.
    root: :class:`bool`
        Whether to set up the root logger rather than the library logger.
    """

    if level is MISSING:
        level = logging.INFO

    if handler is MISSING:
        handler = logging.StreamHandler()

    if formatter is MISSING:
        if isinstance(handler, logging.StreamHandler) and stream_supports_colour(handler.stream):
            formatter = _ColourFormatter()
        else:
            dt_fmt = '%Y-%m-%d %H:%M:%S'
            formatter = logging.Formatter('[{asctime}] [{levelname:<8}] {name}: {message}', dt_fmt, style='{')

    if root:
        logger = logging.getLogger()
    else:
        library, _, _ = __name__.partition('.')
        logger = logging.getLogger(library)

    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)



import datetime
import hashlib
from typing import Optional, Literal, Dict, List, Any

from .utils import format_timestamp
from .http import HTTPClient, UploadPayload
from .models.file import LargeFilePart, PartialFile, File


class LargeFile(PartialFile):
    """Represents a large file being uploaded to Backblaze.

    Attributes
    ----------
    account_id: :class:`str`
        The account's ID that owns the file.
    action: Literal[``'start'``]
        This will always be ``start``.
    bucket_id: :class:`str`
        The file's bucket ID.
    content_type: :class:`str`
        The file's content type, e.g. image/jpeg.
    id: :class:`str`
        The file's ID.
    info: :class:`dict`
        Any custom info regarding the file submitted at upload.
    name: :class:`str`
        The file's name.
    retention: Optional[:class:`dict`]
        The file's object lock retention settings.
    legal_hold: Optional[:class:`dict`]
        The file's object lock legal hold status.
    replication_status: Optional[Literal[``'PENDING'``, ``'COMPLETED'``, ``'FAILED'``, ``'REPLICA'``]
        The file's replication status.
    server_side_encryption: Optional[:class:`dict`]
        The file's encryption mode, and algorithm.
    created: :class:`datetime.datetime`
        The time at which the file was uploaded.
    parts: List[:class:`LargeFilePart`]
        A list of all the uploaded file parts.
    """
    def __init__(
        self,
        data: UploadPayload,
        http: HTTPClient
    ):
        super().__init__(data)

        # appease type checker
        assert data['contentType'] is not None
        assert data['fileId'] is not None

        self.account_id: str = data['accountId']
        self.action: Literal['upload'] = data['action']  # type: ignore
        self.bucket_id: str = data['bucketId']
        self.content_type: str = data['contentType']
        self.info: Dict[Any, Any] = data['fileInfo']
        self.retention: Optional[Dict[Any, Any]] = data.get('fileRetention')
        self.legal_hold: Optional[Dict[Any, Any]] = data.get('legalHold')
        self.replication_status: Optional[Literal['PENDING', 'COMPLETED', 'FAILED', 'REPLICA']] = data.get('replicationStatus')  # noqa
        self.server_side_encryption: Optional[Dict[Any, Any]] = data.get('serverSideEncryption')
        self.created: datetime.datetime = format_timestamp(data['uploadTimestamp'])

        self._http = http
        self._cancelled = False
        self._finished = False
        self._parts: List[LargeFilePart] = []
        self._sha1_checksums: List[str] = []

        self.recommended_part_size: int = self._http._recommended_part_size  # type: ignore
        self.absolute_minimum_part_size: int = self._http._absolute_minimum_part_size  # type: ignore

    async def upload_part(self, content_bytes: bytes) -> LargeFilePart:
        """|coro|

        Uploads a part of the large file.

        Large file parts must be above the ``self.absolute_minimum_part_size``. It's recommended to adhere to the
        ``recommended_part_size`` for the fastest possible upload times.

        Parameters
        ----------
        content_bytes: :class:`bytes`
            The raw bytes of the file part.
        """
        if self._cancelled:
            raise RuntimeError('New parts cannot be uploaded to a cancelled large file upload')
        if self._finished:
            raise RuntimeError('New parts cannot be uploaded to an already complete large file')

        # we could use the Backblaze-sent sha1s instead of storing our own,
        # but this'd defeat the purpose of the final double-check
        sha1 = hashlib.sha1(content_bytes).hexdigest()

        part = await self._http.upload_part(
            self.id,
            len(self._parts) + 1,
            content_bytes,
            sha1
        )
        part = LargeFilePart(part)
        self._parts.append(part)
        self._sha1_checksums.append(sha1)

        return part

    async def finish(self) -> File:
        """|coro|

        Combines every uploaded part into a single file.
        """
        if self._cancelled:
            raise RuntimeError('Cancelled large files cannot be completed')
        if self._finished:
            raise RuntimeError('This large file has already been finished')
        if len(self._parts) <= 1:
            raise RuntimeError('Large files must have at least 2 parts to be finished')

        data = await self._http.finish_large_file(self.id, [i.content_sha1 for i in self._parts])
        return File(data)

    async def cancel(self) -> None:
        """|coro|

        Cancels the upload of a large file, and deletes all of the parts that have been uploaded.
        """
        if self._cancelled:
            raise RuntimeError('This large file upload has already been cancelled')
        if self._finished:
            raise RuntimeError('This large file has already been combined into a single file, it can now only be deleted')

        await self._http.cancel_large_file(self.id)
        self._cancelled = True



from typing import TypedDict, Optional, List


class Permissions(TypedDict):
    capabilities: List[str]
    bucketId: Optional[str]
    bucketName: Optional[str]
    namePrefix: Optional[str]


class AccountAuthorizationPayload(TypedDict):
    accountId: str
    authorizationToken: str
    allowed: Permissions
    apiUrl: str
    downloadUrl: str
    recommendedPartSize: int
    absoluteMinimumPartSize: int
    s3ApiUrl: str
    capabilities: List[str]



from __future__ import annotations

import asyncio
import base64
import datetime
import hashlib
import logging
import re
import sys
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Coroutine,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
    Type,
    Any,
    DefaultDict,
    Dict,
    List,
)
from urllib.parse import quote, quote_plus

import aiohttp
from yarl import URL

from .errors import BackblazeServerError, Forbidden, HTTPException, NotFound, RateLimited, Unauthorized
from .models.account import AccountAuthorizationPayload, Permissions
from .models.file import LargeFilePartPayload, PartialFilePayload, UploadPayload
from .utils import MISSING

if TYPE_CHECKING:
    from typing_extensions import Self
    from types import TracebackType

    BE = TypeVar('BE', bound=BaseException)
    T = TypeVar('T')
    Response = Coroutine[Any, Any, T]

log = logging.getLogger(__name__)


class UploadURLPayload(TypedDict):
    bucketId: str
    uploadUrl: str
    authorizationToken: str


class LargeFileUploadURLPayload(TypedDict):
    fileId: str
    uploadUrl: str
    authorizationToken: str


def handle_upload_file_headers(
    file_name: str,
    content_bytes: bytes,
    content_type: str,
    content_disposition: Optional[str],
    content_language: Optional[List[str]],
    expires: Optional[datetime.datetime],
    content_encoding: Optional[List[Literal['gzip', 'compress', 'deflate', 'identity']]],
    comments: Optional[Dict[str, str]],
    upload_timestamp: Optional[datetime.datetime],
    server_side_encryption: Optional[Literal['AES256']]
) -> Dict[str, Any]:
    headers: Dict[str, Union[str, int]] = {
        'X-Bz-File-Name': quote(file_name),
        'Content-Type': content_type,
        'X-Bz-Content-Sha1': hashlib.sha1(content_bytes).hexdigest()
    }

    if content_disposition is not None:
        reg = re.compile(r'(?P<display>inline|attachment)(?:\s*;\s*filename="(?P<filename>.*)?")?')
        if not reg.search(content_disposition):
            raise ValueError('The Content-Disposition header must be valid')

        headers['X-Bz-Info-b2-content-disposition'] = quote(content_disposition)

    if content_language is not None:
        headers['X-Bz-Info-b2-content-language'] = quote(', '.join(content_language))

    if expires is not None:
        date = expires.astimezone(datetime.timezone.utc).strftime('%a, %d %b %Y %H:%M:%S GMT')
        headers['X-Bz-Info-b2-expires'] = quote(date)

    if content_encoding is not None:
        headers['X-Bz-Info-b2-content-encoding'] = quote(', '.join(content_encoding))

    if comments is not None:
        key, value = list(comments.items())[0]
        headers[f'X-Bz-Info-{quote_plus(key)}'] = quote(value.encode('utf-8'))

    if upload_timestamp is not None:
        if datetime.datetime.now() < upload_timestamp:
            raise ValueError('Future dates for upload timestamps are not supported')
        if upload_timestamp < datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc):
            raise ValueError('The upload timestamp needs to be after January 1st, 1970 UTC')

        timestamp = int(upload_timestamp.timestamp() * 1000)
        headers['X-Bz-Custom-Upload-Timestamp'] = str(timestamp)

    if server_side_encryption is not None:
        if server_side_encryption != 'AES256':
            raise TypeError('Backblaze currently only supports AES256 server-side encryption')

        headers['X-Bz-Server-Side-Encryption'] = server_side_encryption

    return headers


class UploadInfo:
    __slots__ = ('url', 'token', 'created', 'in_use')

    def __init__(self, url: str, token: str, created: datetime.datetime, in_use: bool = False):
        self.url = url
        self.token = token
        self.created = created
        self.in_use = in_use

    @property
    def expires(self) -> datetime.datetime:
        return self.created + datetime.timedelta(days=1)

    def __enter__(self) -> Self:
        self.in_use = True
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BE]],
        exc: Optional[BE],
        traceback: Optional[TracebackType],
    ) -> None:
        self.in_use = False


class BucketUploadInfo(UploadInfo):
    def __repr__(self) -> str:
        return f'<BucketUploadInfo> url={self.url} token={self.token} created={self.created} in_use={self.in_use}'


class LargeFileUploadInfo(UploadInfo):
    def __repr__(self) -> str:
        return f'<LargeFileUploadInfo> url={self.url} token={self.token} created={self.created} in_use={self.in_use}'


class Route:
    """A helper class for instantiating a HTTP method to Backblaze

    Parameters
    -----------
    method: :class:`str`
        The HTTP method you wish to perform, e.g. ``"POST"``
    path: :class:`str`
        The prepended path to the API endpoint you with to target. e.g. ``"/b2_authorize_account"``
    base: Optional[:class:`str`]
        The URL to override the base with. Useful for the b2_download_file_by_name/b2_upload_file Backblaze routes.
    parameters: Any
        This is a special cased kwargs. Anything passed to these will substitute it's key to value in the `path`.
    """

    BASE: str = 'https://api.backblazeb2.com/b2api/v2'

    def __init__(
        self,
        method: Literal['GET', 'POST', 'PUT', 'DELETE'],
        path: str,
        *,
        base: Optional[str] = None,
        **parameters: Any
    ) -> None:
        self.method: Literal['GET', 'POST', 'PUT', 'DELETE'] = method
        self.path = path
        self.parameters = parameters
        url = (base or self.BASE) + self.path
        if parameters:
            url = url.format_map({k: quote(v) if isinstance(v, str) else v for k, v in self.parameters.items()})

        self.url: URL = URL(url, encoded=True)

    def __repr__(self) -> str:
        return f'{self.method} {str(self.url)}'


async def json_or_bytes(response: aiohttp.ClientResponse, route: Route) -> Union[Dict[str, Any], bytes]:
    if response.headers['Content-Type'].lower() in ('application/json;charset=utf-8', 'application/json'):
        return await response.json()
    else:
        return await response.read()


class HTTPClient:
    def __init__(
        self,
        application_key_id: str,
        application_key: str,
        session: Optional[aiohttp.ClientSession] = None
    ):
        self._application_key_id = application_key_id
        self._application_key = application_key
        self._session = session

        self._account_id: str = MISSING
        self._authorization_token: str = MISSING
        self._allowed: Permissions = MISSING
        self._api_url: str = MISSING
        self._download_url: str = MISSING
        self._recommended_part_size: int = MISSING
        self._absolute_minimum_part_size: int = MISSING
        self._s3_api: str = MISSING

        self._upload_urls: DefaultDict[str, List[BucketUploadInfo]] = defaultdict(list)
        self._upload_part_urls: DefaultDict[str, List[LargeFileUploadInfo]] = defaultdict(list)

        # used when first authorizing to update the MISSING headers
        # and when they're refreshed, i.e. when they expire
        self.refresh_headers: bool = True

        # prevent circular import
        from . import __version__
        user_agent = 'aiob2 (https://github.com/Void-ux/aiob2 {0}) Python/{0[0]}.{0[1]} aiohttp/{1}'
        self.user_agent: str = user_agent.format(__version__, sys.version_info, aiohttp.__version__)

    @staticmethod
    async def _generate_session() -> aiohttp.ClientSession:
        """Creates an :class:`aiohttp.ClientSession` for use in the http client.

        Returns
        --------
        :class:`aiohttp.ClientSession`
            The underlying client session we use.

        .. note::
            This method must be a coroutine to avoid the deprecation warning of Python 3.9+.
        """
        return aiohttp.ClientSession()

    async def _close(self) -> None:
        """This method will close the internal client session to ensure a clean exit."""

        if self._session is not None:
            await self._session.close()

    async def _authorize_account(self) -> None:
        """Used to log in to the B2 API."""

        id_and_key = f'{self._application_key_id}:{self._application_key}'.encode()
        basic_auth_string = 'Basic ' + base64.b64encode(id_and_key).decode()
        headers = {
            'Authorization': basic_auth_string
        }

        route = Route('GET', '/b2_authorize_account')
        response: AccountAuthorizationPayload = await self.request(route, headers=headers)

        self._account_id: str = response['accountId']
        self._authorization_token: str = response['authorizationToken']
        self._allowed: Permissions = response['allowed']
        self._api_url: str = response['apiUrl']
        self._download_url: str = response['downloadUrl']
        self._recommended_part_size: int = response['recommendedPartSize']
        self._absolute_minimum_part_size: int = response['absoluteMinimumPartSize']
        self._s3_api: str = response['s3ApiUrl']

    async def _get_upload_url(self, bucket_id: str) -> BucketUploadInfo:
        """Fetches the upload URL and token for a specific bucket or a large file.

        Parameters
        -----------
        bucket_id: :class:`str`
            The ID of the bucket to get the upload info for.

        Returns
        ---------
        :class:`BucketUploadInfo`
            The URL for uploading files or file parts, and the authorization token to use with it.
        """

        headers = {
            'Authorization': self._authorization_token
        }
        params = {
            'bucketId': bucket_id
        }

        now = datetime.datetime.now()
        route = Route('GET', '/b2api/v2/b2_get_upload_url', base=self._api_url)
        response: UploadURLPayload = await self.request(route, headers=headers, params=params)
        return BucketUploadInfo(response['uploadUrl'], response['authorizationToken'], now)

    async def start_large_file(
        self,
        bucket_id: str,
        file_name: str,
        content_type: str,
        upload_timestamp: Optional[datetime.datetime],
        comments: Optional[Dict[str, str]]
    ) -> UploadPayload:
        """Prepares for uploading every part of a large file."""

        headers = {
            'Authorization': self._authorization_token
        }
        data: Dict[Any, Any] = {
            'bucketId': bucket_id,
            'fileName': file_name,
            'contentType': content_type
        }

        if upload_timestamp:
            assert datetime.datetime.now() >= upload_timestamp, 'Future dates for upload timestamps are not supported'

            timestamp = int(upload_timestamp.timestamp() * 1000)
            assert timestamp.bit_length() <= 64, \
                'The upload timestamp must be 64 bits when turned into a UNIX timestamp in milliseconds'

            headers['X-Bz-Custom-Upload-Timestamp'] = str(timestamp)
        if comments:
            data['fileInfo'] = comments

        # Backblaze allows future dates for some accounts,
        # so pre-checking it isn't viable
        if upload_timestamp:
            timestamp = int(upload_timestamp.timestamp() * 1000)
            assert timestamp.bit_length() <= 64, \
                'The upload timestamp must be 64 bits when turned into a UNIX timestamp in milliseconds.'
            data['X-Bz-Custom-Upload-Timestamp'] = str(timestamp)

        route = Route('POST', '/b2api/v2/b2_start_large_file', base=self._api_url)
        return await self.request(route, json=data, headers=headers)

    async def _get_upload_part_url(self, file_id: str) -> LargeFileUploadInfo:
        """Fetches an upload URL and token for uploading parts of a large file.

        Parameters
        -----------
        file_id: :class:`str`
            The ID of the large file.

        Returns
        ---------
        :class:`LargeFileUploadInfo`
            The URL for uploading files, and the authorization token to use with it.
        """

        headers = {
            'Authorization': self._authorization_token
        }
        params = {
            'fileId': file_id
        }

        now = datetime.datetime.now()
        route = Route('GET', '/b2api/v2/b2_get_upload_part_url', base=self._api_url)
        response: LargeFileUploadURLPayload = await self.request(route, headers=headers, params=params)
        return LargeFileUploadInfo(response['uploadUrl'], response['authorizationToken'], now)

    async def request(
        self,
        route: Route,
        *,
        bucket_id: Optional[str] = None,
        large_file_id: Optional[str] = None,
        upload_info: Optional[Union[BucketUploadInfo, LargeFileUploadInfo]] = None,
        **kwargs: Any
    ) -> Any:
        """Send a HTTP request to the `route.url`, with HTTP error code handling defined by the B2 API docs.

        Parameters
        ----------
        route: :class:`Route`
            The HTTP method and URL.
        bucket_id: Optional[:class:`str`]
            The bucket ID being uploaded to, only specified when called from `HTTPClient.upload_file`.
        **kwargs: Any
            kwargs to pass to `aiohttp.ClientSession.request()`, any of `headers`, `params`, `data`, `json`.
        """

        # we'll use this variable to tell whether or not to refresh an upload
        # url url/token, or the account's api url/token. this'll only be `True`
        # if we're using b2_upload_file or b2_upload_part
        uploading_file = upload_info is not None

        if self._session is None:
            self._session = await self._generate_session()

        headers: Dict[str, str] = kwargs.pop('headers')
        headers['User-Agent'] = self.user_agent

        # this will only ever happen on the very first self.request
        # luckily, the first route will always use the account authorization
        # token, e.g. b2_get_upload_url, download_file_by_x, etc.
        # only b2_upload_file and b2_upload_part use custom tokens, which must first
        # go through b2_get_upload_(part_)url, so they can't be the first request.
        if headers['Authorization'] is MISSING and route.path != '/b2_authorize_account':
            log.info('Authenticating using static application key')
            await self._authorize_account()
            log.debug('Logged in with account ID %s', self._account_id)
            headers['Authorization'] = self._authorization_token
            route = Route(route.method, route.path, base=self._api_url, **route.parameters)

        for tries in range(5):
            if upload_info:
                upload_info.in_use = True

            async with self._session.request(route.method, route.url, headers=headers, **kwargs) as response:
                if upload_info:
                    upload_info.in_use = False

                if uploading_file:
                    log.debug(
                        '%s %s with a payload of %s bytes has returned %s',
                        route.method,
                        route.url,
                        len(kwargs['data']),
                        response.status
                    )
                else:
                    log.debug('%s %s has returned %s', route.method, route.url, response.status)

                data = await json_or_bytes(response, route)

                if 300 > response.status >= 200:
                    # for download_file_by_x; the headers contain info about the files
                    if isinstance(data, bytes):
                        return data, response.headers
                    return data

                # appease type checker; json_or_bytes will
                # only return bytes if response.ok which'll
                # be handled in the `if` statement above
                assert isinstance(data, dict)

                # we are being rate limited
                if response.status == 429:
                    raise RateLimited(response, data)

                if response.status in {408, 500}:
                    log.warning(
                        '%s %s returned %s (%s), sleeping for %s',
                        route.method,
                        route.url,
                        response.status,
                        data['code'],
                        1 + tries * 2
                    )
                    await asyncio.sleep(1 + tries * 2)
                    continue

                # Backblaze recommends calling b2_get_upload_(part_)url again with 503s
                if response.status == 401 or (response.status == 503 and uploading_file):
                    if route.path == '/b2_authorize_account':
                        raise Unauthorized(response, data)

                    if data['code'] not in (
                        'expired_auth_token',  # auth token has expired (created over 24 hours ago)
                        'bad_auth_token',      # auth token was not valid to begin with
                        'service_unavailable'  # internal server error i.e. 503
                    ):
                        raise Unauthorized(response, data)

                    if uploading_file and upload_info is not None:
                        # we need to remove this token to prevent it from being
                        # used again, since find_upload_(part_)url always returns
                        # the first element found. this is only for edge-cases when
                        # Backblaze actually invalidates the token, since aiob2 will
                        # prevent expired tokens from being used.

                        if bucket_id is not None and isinstance(upload_info, BucketUploadInfo):
                            old_upload_info = \
                                [i for i in self._upload_urls[bucket_id] if i.token == headers['Authorization']]
                            self._upload_urls[bucket_id].remove(upload_info)
                            log.debug('Removed an expired/invalid upload URL for bucket ID %s', bucket_id)

                            upload_info = await self._find_upload_url(bucket_id)
                        elif large_file_id is not None and isinstance(upload_info, LargeFileUploadInfo):
                            old_upload_info = \
                                [i for i in self._upload_part_urls[large_file_id] if i.token == headers['Authorization']]
                            self._upload_part_urls[large_file_id].remove(old_upload_info[0])
                            log.debug('Removed an expired/invalid upload URL for large file ID %s', large_file_id)

                            upload_info = await self._find_upload_part_url(large_file_id)

                        # reset the upload URL and token and retry
                        headers['Authorization'] = upload_info.token
                        route = Route(
                            route.method,
                            route.path,
                            base=upload_info.url,
                            parameters=route.parameters
                        )

                        log.info('Re-authenticated upload URL and upload URL token')
                    else:
                        # download_by_file_x also uses the account authorization token
                        # so we don't need a seperate elif
                        await self._authorize_account()
                        headers['Authorization'] = self._authorization_token

                        log.info('Re-authenticated account authorization token')

                    continue

                if response.status == 403:
                    raise Forbidden(response, data)
                elif response.status == 404:
                    raise NotFound(response, data)
                elif response.status >= 500:
                    raise BackblazeServerError(response, data)
                else:
                    raise HTTPException(response, data)

    def _purge_expired_upload_urls(self, bucket_id: str) -> None:
        old = self._upload_urls[bucket_id]
        new = self._upload_urls[bucket_id] = list(filter(
            lambda x: x.expires > datetime.datetime.now(),
            self._upload_urls[bucket_id]
        ))

        if len(old) > len(new):
            log.debug('Purged %s expired upload URLs for bucket ID %s', len(old) - len(new), bucket_id)

    async def _find_upload_url(self, bucket_id: str) -> BucketUploadInfo:
        """Attempts to find an existing non-expired, not-in-use upload URL for a bucket, otherwise it'll fetch a new one.

        Note
        ~~~~
            This will also purge all the expired upload URLs from `self._upload_urls[bucket_id]` and append the newly-created
            one if applicable.

        Parameters
        ----------
        bucket_id: str
            The bucket_id to find/create an upload URL for.

        Returns
        -------
        :class:`BucketUploadInfo`
            The resolved bucket upload info, or the newly-created one.
        """
        self._purge_expired_upload_urls(bucket_id)
        upload_infos = self._upload_urls[bucket_id]

        if not upload_infos:
            log.debug('No existing upload URLs found for bucket ID %s, creating a new one', bucket_id)
            upload_info = await self._get_upload_url(bucket_id)
            self._upload_urls[bucket_id].append(upload_info)
        else:
            if all(x.in_use for x in upload_infos):
                log.debug('All existing upload URLs for bucket ID %s are in-use, creating a new one', bucket_id)
                upload_info = await self._get_upload_url(bucket_id)
                self._upload_urls[bucket_id].append(upload_info)
            else:
                log.debug('Using existing upload URL for bucket ID %s', bucket_id)
                upload_info = list(filter(lambda x: not x.in_use, upload_infos))[0]

        return upload_info

    async def upload_file(
        self,
        *,
        file_name: str,
        content_bytes: bytes,
        bucket_id: str,
        content_type: str,
        content_disposition: Optional[str],
        content_language: Optional[List[str]],
        expires: Optional[datetime.datetime],
        content_encoding: Optional[List[Literal['gzip', 'compress', 'deflate', 'identity']]],
        comments: Optional[Dict[str, str]],
        upload_timestamp: Optional[datetime.datetime],
        server_side_encryption: Optional[Literal['AES256']]
    ) -> UploadPayload:
        upload_info = await self._find_upload_url(bucket_id)

        headers = handle_upload_file_headers(
            file_name,
            content_bytes,
            content_type,
            content_disposition,
            content_language,
            expires,
            content_encoding,
            comments,
            upload_timestamp,
            server_side_encryption
        )
        headers['Authorization'] = upload_info.token

        route = Route('POST', '', base=upload_info.url)
        return await self.request(route, bucket_id=bucket_id, upload_info=upload_info, headers=headers, data=content_bytes)

    def _purge_expired_upload_part_urls(self, large_file_id: str) -> None:
        old = self._upload_part_urls[large_file_id]
        new = self._upload_part_urls[large_file_id] = list(filter(
            lambda x: x.expires > datetime.datetime.now(),
            self._upload_part_urls[large_file_id]
        ))

        if len(old) > len(new):
            log.debug('Purged %s expired upload URLs for large file ID %s', len(old) - len(new), large_file_id)

    async def _find_upload_part_url(self, large_file_id: str) -> LargeFileUploadInfo:
        """Attempts to find an existing non-expired, not-in-use large upload URL for a bucket, otherwise it'll fetch a new
        one.

        Note
        ~~~~
            This will also purge all the expired large upload URLs from `self._upload_urls[bucket_id]` and append the
            newly-created one if applicable.

        Parameters
        ----------
        bucket_id: str
            The bucket_id to find/create a large upload URL for.

        Returns
        -------
        :class:`LargeFileUploadInfo`
            The resolved bucket upload info, or the newly-created one.
        """
        self._purge_expired_upload_part_urls(large_file_id)
        upload_part_infos = self._upload_part_urls.get(large_file_id)

        # remember, upload_part_infos could also be []
        if not upload_part_infos:
            log.debug('No existing upload URLs found for large file ID %s, creating a new one', large_file_id)
            upload_part_info = await self._get_upload_part_url(large_file_id)
            self._upload_part_urls[large_file_id].append(upload_part_info)
        else:
            self._purge_expired_upload_part_urls(large_file_id)

            if all(x.in_use for x in upload_part_infos):
                log.debug('All existing upload URLs for large file ID %s are in-use, creating a new one', large_file_id)
                upload_part_info = await self._get_upload_part_url(large_file_id)
                self._upload_part_urls[large_file_id].append(upload_part_info)
            else:
                log.debug('Using existing upload URL for large file ID %s', large_file_id)
                upload_part_info = list(filter(lambda x: not x.in_use, upload_part_infos))[0]

        return upload_part_info

    async def upload_part(
        self,
        file_id: str,
        part_number: int,
        content_bytes: bytes,
        sha1: str
    ) -> LargeFilePartPayload:
        upload_info = await self._find_upload_part_url(file_id)

        content_length = len(content_bytes)
        if content_length > self._recommended_part_size:
            log.warning(
                'Upload part %s for file ID %s is over the recommended part size, this may result in longer upload times',
                part_number,
                file_id
            )

        headers = {
            'Authorization': upload_info.token,
            'X-Bz-Part-Number': str(part_number),
            'Content-Length': str(content_length),
            'X-Bz-Content-Sha1': sha1
        }

        route = Route('POST', '', base=upload_info.url)
        return await self.request(route, large_file_id=file_id, headers=headers, data=content_bytes)

    def finish_large_file(self, file_id: str, sha1_list: list[str]) -> Response[UploadPayload]:
        headers = {
            'Authorization': self._authorization_token
        }
        data = {
            'fileId': file_id,
            'partSha1Array': sha1_list
        }

        route = Route('POST', '/b2api/v2/b2_finish_large_file', base=self._api_url)
        return self.request(route, headers=headers, json=data)

    def cancel_large_file(self, file_id: str) -> Response[None]:
        headers = {
            'Authorization': self._authorization_token
        }
        data = {
            'fileId': file_id
        }

        route = Route('POST', '/b2api/v2/b2_cancel_large_file', base=self._api_url)
        return self.request(route, headers=headers, json=data)

    def delete_file(self, *, file_name: str, file_id: str) -> Response[PartialFilePayload]:
        route = Route('POST', '/b2api/v2/b2_delete_file_version', base=self._api_url)
        headers = {
            'Authorization': self._authorization_token
        }
        data = {
            'fileName': file_name,
            'fileId': file_id
        }
        return self.request(route, headers=headers, json=data)

    def download_file_by_id(
        self,
        *,
        file_id: str,
        content_disposition: Optional[str] = None,
        content_language: Optional[str] = None,
        expires: Optional[str] = None,
        cache_control: Optional[str] = None,
        content_encoding: Optional[str] = None,
        content_type: Optional[str] = None,
        server_side_encryption: Optional[str] = None
    ) -> Response[Tuple[bytes, Dict[str, Any]]]:
        headers = {
            'Authorization': self._authorization_token,
            'b2ContentDisposition': content_disposition,
            'b2ContentLanguage': content_language,
            'b2Expires': expires,
            'b2CacheControl': cache_control,
            'b2ContentEncoding': content_encoding,
            'b2ContentType': content_type,
            'serverSideEncryption': server_side_encryption
        }
        headers = {key: value for key, value in headers.items() if value is not None}
        params = {
            'fileId': file_id
        }
        route = Route(
            'GET',
            '/b2api/v2/b2_download_file_by_id',
            base=self._download_url
        )

        return self.request(route, headers=headers, params=params)

    def download_file_by_name(
        self,
        *,
        file_name: str,
        bucket_name: str,
        content_disposition: Optional[str] = None,
        content_language: Optional[str] = None,
        expires: Optional[str] = None,
        cache_control: Optional[str] = None,
        content_encoding: Optional[str] = None,
        content_type: Optional[str] = None,
        server_side_encryption: Optional[str] = None
    ) -> Response[Tuple[bytes, Dict[str, Any]]]:
        headers = {
            'Authorization': self._authorization_token,
            'b2ContentDisposition': content_disposition,
            'b2ContentLanguage': content_language,
            'b2Expires': expires,
            'b2CacheControl': cache_control,
            'b2ContentEncoding': content_encoding,
            'b2ContentType': content_type,
            'serverSideEncryption': server_side_encryption
        }
        headers = {key: value for key, value in headers.items() if value is not None}
        route = Route(
            'GET',
            '/file/{bucket_name}/{file_name}',
            base=self._download_url,
            bucket_name=bucket_name,
            file_name=file_name
        )

        return self.request(route, headers=headers)



from __future__ import annotations

import datetime
from typing import TypedDict, Literal, Optional, Union, Dict, Any
from typing_extensions import NotRequired

from .archetypes import B2Object
from ..utils import format_timestamp

__all__ = ('File', 'DeletedFile', 'DownloadedFile')


class PartialFilePayload(TypedDict):
    fileName: str
    fileId: Optional[str]


class UploadPayload(PartialFilePayload):
    accountId: str
    action: Literal['start', 'upload', 'hide', 'folder']
    bucketId: str
    contentLength: Optional[int]
    contentSha1: Optional[str]
    contentMd5: NotRequired[Optional[str]]
    contentType: Optional[str]
    fileInfo: Dict[Any, Any]
    fileRetention: NotRequired[Optional[Dict[Any, Any]]]
    legalHold: NotRequired[Dict[Any, Any]]
    replicationStatus: NotRequired[Literal['PENDING', 'COMPLETED', 'FAILED', 'REPLICA']]
    serverSideEncryption: NotRequired[Optional[Dict[Any, Any]]]
    uploadTimestamp: Union[int, Literal[0]]


DownloadPayloadHeaders = TypedDict('DownloadPayloadHeaders', {
    'Content-Length': int,
    'Content-Type': str,
    'X-Bz-File-Id': str,
    'X-Bz-File-Name': str,
    'X-Bz-Content-Sha1': str,
    'X-Bz-Upload-Timestamp': str,
    'Accept-Ranges': str,
    'Date': str,
    'Content-Disposition': NotRequired[str],
    'Content-Language': NotRequired[str],
    'Expires': NotRequired[str],
    'Content-Encoding': NotRequired[str],
    'X-Bz-Server-Side-Encryption': NotRequired[Literal['AES256']]
})


class PartialFile(B2Object):
    """Represents a "partial" file from Backblaze.

    Attributes
    ----------
    name: :class:`str`
        The file's name.
    id: :class:`str`
        The file's ID.
    """
    def __init__(
        self,
        data: PartialFilePayload
    ):
        # appease type checker
        assert data['fileId'] is not None

        self.name: str = data['fileName']
        self.id: str = data['fileId']

    def __str__(self):
        return self.name

    def __eq__(self, other: Any):
        return isinstance(other, File) and self.id == other.id


class LargeFilePartPayload(TypedDict):
    fileId: str
    partNumber: int
    contentLength: int
    contentSha1: str
    contentMd5: Optional[str]
    serverSideEncryption: Optional[Dict[Any, Any]]
    uploadTimestamp: int


class LargeFilePart(B2Object):
    def __init__(self, payload: LargeFilePartPayload) -> None:
        self.file_id: str = payload['fileId']
        self.part_number: int = payload['partNumber']
        self.content_length: int = payload['contentLength']
        self.content_sha1: str = payload['contentSha1']
        self.content_md5: Optional[str] = payload['contentMd5']
        self.server_side_encryption: Optional[Dict[Any, Any]] = payload['serverSideEncryption']
        self.upload_timestamp: datetime.datetime = format_timestamp(payload['uploadTimestamp'])


class File(PartialFile):
    """Represents a file uploaded to Backblaze.

    Attributes
    ----------
    account_id: :class:`str`
        The account's ID that owns the file.
    action: Literal[``'upload'``]
        This will always be ``upload``.
    bucket_id: :class:`str`
        The file's bucket ID.
    content_length: :class:`int`
        The file's size represented in number of bytes.
    content_sha1: :class:`str`
        The file's SHA-1.
    content_md5: Optional[:class:`str`]
        The MD5 of the file's bytes as a 40-digit hex string.
    content_type: :class:`str`
        The file's content type, e.g. image/jpeg.
    id: :class:`str`
        The file's ID.
    info: :class:`dict`
        Any custom info regarding the file submitted at upload.
    name: :class:`str`
        The file's name.
    retention: Optional[:class:`dict`]
        The file's object lock retention settings.
    legal_hold: Optional[:class:`dict`]
        The file's object lock legal hold status.
    replication_status: Optional[Literal[``'PENDING'``, ``'COMPLETED'``, ``'FAILED'``, ``'REPLICA'``]
        The file's replication status.
    server_side_encryption: Optional[:class:`dict`]
        The file's encryption mode, and algorithm.
    created: :class:`datetime.datetime`
        The time at which the file was uploaded.
    """
    def __init__(
        self,
        data: UploadPayload
    ):
        super().__init__(data)

        # appease type checker
        assert data['contentLength'] is not None
        assert data['contentSha1'] is not None
        assert data['contentType'] is not None

        self.account_id: str = data['accountId']
        self.action: Literal['upload'] = data['action']  # type: ignore
        self.bucket_id: str = data['bucketId']
        self.content_length: int = data['contentLength']
        self.content_sha1: str = data['contentSha1']
        self.content_md5: Optional[str] = data.get('contentMd5')
        self.content_type: str = data['contentType']
        self.info: Dict[Any, Any] = data['fileInfo']
        self.retention: Optional[Dict[Any, Any]] = data.get('fileRetention')
        self.legal_hold: Optional[Dict[Any, Any]] = data.get('legalHold')
        self.replication_status: Optional[Literal['PENDING', 'COMPLETED', 'FAILED', 'REPLICA']] = data.get('replicationStatus')  # noqa
        self.server_side_encryption: Optional[Dict[Any, Any]] = data.get('serverSideEncryption')
        self.created: datetime.datetime = format_timestamp(data['uploadTimestamp'])


class DeletedFile(PartialFile):
    """Represents a deleted file from Backblaze.

    Attributes
    ----------
    name: :class:`str`
        The file's name.
    id: :class:`str`
        The file's ID.
    """
    def __init__(
        self,
        data: PartialFilePayload
    ):
        super().__init__(data)


class DownloadedFile(B2Object):
    """Represents a file downloaded from Backblaze.

    Attributes
    ----------
    name: :class:`str`
        The file's name.
    id: :class:`str`
        The file's ID.
    content_type: :class:`str`
        The file's content type, e.g. image/jpeg.
    content_length: :class:`int`
        The file's size represented in number of bytes.
    content_sha1: :class:`str`
        The file's SHA-1.
    created: :class:`datetime.datetime`
        The time at which the file was uploaded.
    downloaded_at: :class:`str`
        The date at which the download was requested.
    content_disposition: Optional[:class:`str`]
        Whether or not the content is intended to be played inline, or downloaded locally and optionally the filename.
    content_language: Optional[:class:`str`]
        The content's intended language audience.
    expires: Optional[:class:`datetime.datetime`]
        The intended expiration date of the content.
    content_encoding: Optional[:class:`str`]
        The content's encoding in the order they were performed in.
    server_side_encryption: Optional[Literal[``AES256``]]
        The server side encryption performed on the content including the algorithm.
    comments: Optional[Dict[:class:`str`, :class:`str`]]
        The comments uploaded with the file.
    content: :class:`bytes`
        The raw bytes of the downloaded file.
    """
    def __init__(
        self,
        content: bytes,
        headers_: DownloadPayloadHeaders
    ):
        headers: Dict[str, Any] = {k.lower(): v for k, v in headers_.items()}
        b2_info_headers = (
            'b2-content-disposition',
            'b2-content-language',
            'b2-expires',
            'b2-cache-control',
            'b2-content-encoding',
            'b2-content-type'
        )
        # comments are turned to lower case by Backblaze
        # so the above lowercasing won't impact anything
        comments: Dict[str, str] = {
            k[10:]: v for k, v in headers.items() if k.startswith('x-bz-info-') and not k.endswith(b2_info_headers)
        }

        self.content_length: int = headers['content-length']
        self.content_type: str = headers['content-type']
        self.id: str = headers['x-bz-file-id']
        self.name: str = headers['x-bz-file-name']
        self.content_sha1: str = headers['x-bz-content-sha1']
        self.created: datetime.datetime = format_timestamp(int(headers['x-bz-upload-timestamp']))
        self.downloaded_at: datetime.datetime = datetime.datetime.strptime(
            headers['date'],
            '%a, %d %b %Y %H:%M:%S %Z'
        )

        self.content_disposition: Optional[str] = headers.get('content-disposition')
        self.content_language: Optional[str] = headers.get('content-language')
        if (expires := headers.get('cache-control')) is not None:
            expires = datetime.datetime.strptime(expires, '%a, %d %b %Y %H:%M:%S %Z')
        else:
            expires = None
        self.expires: Optional[datetime.datetime] = expires
        self.content_encoding: Optional[str] = headers.get('content-encoding')
        self.server_side_encryption: Optional[Literal['AES256']] = headers.get('x-bz-server-side-encryption')
        self.comments: Optional[Dict[str, str]] = comments or None

        self.content: bytes = content



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiob2', 'aiob2.models']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0', 'typing-extensions>=4.4.0,<5.0.0']

setup_kwargs = {
    'name': 'aiob2',
    'version': '0.8.2',
    'description': "A simple and easy to use async wrapper for Backblaze's B2 bucket API.",
    'long_description': '# aiob2\n\n---\n\n<p align="center">\n    <a href="https://www.python.org/downloads/">\n        <img src="https://img.shields.io/pypi/pyversions/aiob2?style=for-the-badge" alt="Python version">\n    </a>\n    <a href="https://github.com/Void-ux/aiob2/actions">\n        <img src="https://img.shields.io/github/actions/workflow/status/Void-ux/aiob2/build.yaml?branch=master&style=for-the-badge" alt="Build status">\n    </a>\n    <a href="https://pypi.org/project/aiob2/">\n        <img src="https://img.shields.io/pypi/v/aiob2?color=8BC34A&style=for-the-badge" alt="PyPi">\n    </a>\n    <a href="https://opensource.org/licenses/MIT">\n        <img src="https://img.shields.io/pypi/l/aiob2?color=C0C0C0&style=for-the-badge" alt="License">\n    </a>\n</p>\n\naiob2 is an asynchronous API wrapper for the [Backblaze B2 Bucket API](https://www.backblaze.com/b2/docs/calling.html).\n\nIt will allow you to interact with your B2 bucket, it\'s files and anything else that the B2 API allows in a modern, object-oriented fashion.\n\n**NOTE:** This API wrapper is by no means *complete* and has many endpoints to cover, though the main ones have been covered (they will be listed below)\n\n## Installation\n\n---\n\naiob2 is compatible with Python 3.8+ (this is an estimate). To install aiob2, run the following command in your (virtual) environment.\n\n```shell\npip install aiob2\n```\n\nAlternatively, for the latest though least stable version, you can download it from the GitHub repo:\n\n```shell\npip install git+https://github.com/Void-ux/aiob2.git\n```\n\n## Usage\n\n### Uploading\n\n```python\nimport aiohttp\nimport asyncio\n\nfrom aiob2 import Client\n\n# Our image to upload to our bucket\nwith open(r\'C:\\Users\\MS1\\Pictures\\Camera Roll\\IMG_5316.jpeg\', \'rb\') as file:\n    data = file.read()\n\nasync def main():\n    async with Client(\'key_id\', \'key\') as client:\n        file = await client.upload_file(\n            content_bytes=data,\n            file_name=\'test.jpg\',\n            bucket_id=\'bucket_id\',\n        )\n\n\nif __name__ == \'__main__\':\n    asyncio.run(main())\n```\n\nAnd that\'s it! `upload_file()` returns a `File` object that neatly wraps everything Backblaze\'s API has provided us with.\nThe `File` object\'s documentation can be found [here](https://aiob2.readthedocs.io/en/latest/pages/api.html#aiob2.File)\n\n## License\n\n---\n\nThis project is released under the [MIT License](https://opensource.org/licenses/MIT).\n',
    'author': 'Dan',
    'author_email': 'the.void.altacc@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Void-ux/aiob2',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
